***********************************************************************************
      program boost
***********************************************************************************
* To distinguish signal and background by using ada(epsilon)-boosting algorithms
* ---------------------------------------------------------------------------------
* Many thanks to Prof. Ji Zhu for introducing the boosting algorithms and
* to Prof. Byron P. Roe for various cross checking.
*
* References:
* =============
* 1. Ji Zhu, Hai-Jun Yang, Byron P. Roe(2004), "nue and pi0 Separation in the
*    MiniBooNE Experiment by using the Boosting Algorithm", BooNE-TN-112, Jan. 9, 2004
*    http://tenaya.physics.lsa.umich.edu/~hyang/talks/boone-tn112.pdf
*
* 2. Hai-Jun Yang, Byron P. Roe, Ji Zhu(2004), "Particle Identification in the MiniBooNE
*    Experiment by using the Boosting Algorithm", BooNE-TN-113, Feb. 23, 2004
*    http://tenaya.physics.lsa.umich.edu/~hyang/talks/boone-tn113.pdf
*
* 3. Y. Freund and R. E. Schapire(1996), "Experiments with a new boosting algorithm"
*    Proc COLT, 209-217, ACM Press, New York.
*
* 4. J. Friedman(2001), "Greedy function approximation: a gradient boosting machine."
*    Annals of Statistics, 29:5.
*
* 5. Byron P. Roe, Hai-Jun Yang, Ji Zhu, Yong Liu, Ion Stancu, Gordon McGregor,
*    "Boosted Decision Trees, an Alternative to Artificial Neural Networks for Particle
*     Identification", arXiv e-print:physics/0408124, NIMA(2005) in press.
* ---------------------------------------------------------------------------------
* Created by Haijun Yang @ University of Michigan on February 25, 2004.
* Any questions, suggestions, comments and bugs, please send mail to
* yhj@umich.edu, or call 1-734-764-3407.
* ---------------------------------------------------------------------------------
* Version 1.0 released on February 25, 2004 by Haijun Yang
*
* ---------------------------------------------------------------------------------
* Update information should be listed in the following;
* ---------------------------------------------------------------------------------
* 1. Add reference 5, December 10, 2004
*   ....
* ---------------------------------------------------------------------------------
* Brief description of Boosting Algorithms
* =========================================
* Notations: trainning data(x_1,y_1),(x_2,y_2),...,(x_n,y_n), where x_i is real,
*            y_i = [-1,1] (-1 for background, 1 for signal), 
*            classifier T_m(x): x -> [-1,1]
*
* 1. AdaBoost algorithm
* ======================
*   step 1: Initialize the observation weights wt_i = 1. i=1,2,...,n
*   step 2: For m = 1 to M ! number of binary trees
*          2.a) Fit a classifier T_m(x) to the trainning data using weights wt_i
*          2.b) Compute err_m = sum(wt_i*I(y_i.NE.T_m(x_i))/sum(wt_i), i = 1,2,...,n
*                              I = 0 if y_i = T_m(x_i); I = 1 if y_i.NE.T_m(x_i)
*          2.c) Compute alpha_m = beta*log((1.-err_m)/err_m)
*          2.d) Set wt_i = wt_i*exp(alpha_m*I(y_i.NE.T_m(x_i))), i = 1,2,...,n
*          2.e) Re-normalize wt_i = wt_i*n/sum(wt_i)
*   step 3: Output T(x) = sum(alpha_m*T_m(x))
*
* 2. epsilon-boosting algorithm
* =============================
*   step 1: Initialize the observation weights wt_i = 1. i=1,2,...,n
*   step 2: For m = 1 to M ! number of binary trees
*          2.a) Fit a classifier T_m(x) to the trainning data using weights wt_i
*          2.b) Set wt_i = wt_i*exp(2*epsilon*I(y_i.NE.T_m(x_i))), i = 1,2,...,n
*          2.c) Re-normalize wt_i = wt_i*n/sum(wt_i)
*   step 3: Output T(x) = sum(epsilon*T_m(x))
* ---------------------------------------------------------------------------------
* All subroutines used in program
*   call vzero        - from CERN lib
*   call ranmar       - uniform random generator(0,1) from CERN lib
*   call plot         - fill histograms/ntuples if needed
*   call readin_dat   - input dat (events are selected randomly for trainning)
*   call tree_creator - create one binary tree(gini index is used as criteria for next split) 
*   call tree_creator2- create one binary tree(no gini index is used, leaves # varies)
*   call var_min_max  - determine min/max values of each variable
*   call event_bin    - fill events into bins with uniform step size
*   call node_finder  - try to find subnodes of current node
*   call check_purity - calculate test results for existing trees, called every 10 trees
*   call check_variable-calculate # of times used as splitter for each variable
*   call save_leafindex - save trainned classifier trees information (Yang's format)
*   call save_treeindex - save trainned classifier trees information (Roe's format - default)
* --------------------------------------------------------------------------------- 
* Parameters are listed in boost.inc file
* ---------------------------------------------------------------------------------
***********************************************************************************
      include 'boost.inc'
      include 'hbookboost.inc'

      real subindex(node,12)
      real tree_subindex(ntr,node,12)
      real leaf_index(node/2,2*node) ! # of leaves = (# of nodes + 1)/2

      integer tree_index(nevt)
      real tree_combined(nevt)
      real gini_index_all(ntr,node)
      integer itr

      call vzero(tree_combined,nevt)
      call vzero(tree_subindex,ntr*node*12)
      call vzero(gini_index_all,ntr*node)

* open ntp file
      CALL HLIMIT(NWPAWC)

      CALL PLOT(1)

*--- init
      do i = 1, ntrain
         wt(i) = 1.
      enddo

      idebug = 0
      call vzero(alpha,ntr)
      alpha_tot = 0.

      if(iepsilon.eq.1) then
         print *, 'epsilon boosting algorithm is used: ',epsilon
      else
         print *, 'ada boosting algorithm is used: ',beta
      endif

*--- readin data
      call readin_dat(datnode,idebug)
      call save_leafindex(0,leaf_index) ! Yang's format

      do itr = 1, ntr

*--- create one binary tree (Gini_index is used for next split)
         if(ngini_index.eq.1) then
            print *, 'Gini index is used as criteria for next split'
            call tree_creator(datnode,subindex,leaf_index)
         else
            print *, 'Left leave is always tried first for next split'
            call tree_creator2(datnode,subindex,leaf_index)
         endif

         print *, 'ntot/nevt1/2/nleaf = ', ntot, nevt1, nevt2, nleaf

*--- save gini_index
         do i = 1, node-1
            gini_index_all(itr,i) = gini_index(i)
         enddo

*--- save subindex for writeout
         do i = 1, node-1
            do j = 1, 12
               tree_subindex(itr,i,j) = subindex(i,j)
            enddo
         enddo
         
*--- recover datnode
         do i = 1, ntrain
            do j = 1, nvar            
               datnode(j,i) = dat_train(j,i)
            enddo
         enddo

         call vzero(tree_index,nevt)

*--- test binary tree 
         do i = 1, nevt_test
            do ileaf = 1, nleaf
               icheck = 0
               do jcut = 1, leaf_index(ileaf,1)
                  vnum = leaf_index(ileaf,3*jcut+1)
                  vcut = leaf_index(ileaf,3*jcut+2)
                  vlr  = leaf_index(ileaf,3*jcut+3)
                  
                  if(vlr.gt.998.and.dat_test(vnum,i).ge.vcut) then ! right node (great than vcut)
                     icheck = icheck + 1
                  elseif(vlr.lt.-998.and.dat_test(vnum,i).lt.vcut) then ! left node (less than vcut)
                     icheck = icheck + 1
                  endif
               enddo

               if(icheck.eq.leaf_index(ileaf,1)) then
                  if(leaf_index(ileaf,2)/leaf_index(ileaf,3)
     &                 .gt.psigout) then
                     tree_index(i) = 1
                  else
                     tree_index(i) = -1
                  endif                 
               endif
            enddo
         enddo


*--- update weight for next binary tree
         call update_weight(itr,leaf_index)

*--- combine all tree information and check purity of existing trees
         if(iepsilon.eq.1) then ! epsilon boosting algorithm
            do i = 1, nevt_test
               tree_combined(i)=tree_combined(i)+tree_index(i)
            enddo
         else                   ! adaboosting algorithm
            do i = 1, nevt_test
              tree_combined(i)=tree_combined(i)+tree_index(i)*alpha(itr)
            enddo
            alpha_tot = alpha_tot + alpha(itr)
         endif
         

         if(mod(itr,10).eq.0) then
            call check_purity(itr,tree_combined)
         endif

         call save_leafindex(1,leaf_index) ! Yang's format

*--- binary tree loop end here
      enddo                     ! do itr = 1, ntr

*---> check irrelevant variables

      call check_variable(tree_subindex,gini_index_all)

*---> close ntuple

      CALL PLOT(3)

*---> write treeindex.dat file for future test
      
      call save_treeindex(tree_subindex) ! Roe's format
      call save_leafindex(2,leaf_index) ! Yang's format

*                                                           end boost
      end
****************************************************************************
      subroutine save_leafindex(iflag,leaf_index) 
* save all leaves information needed to test new MC 
* Yang's format
****************************************************************************
      include 'boost.inc'
      real leaf_index(node/2,2*node) ! # of leaves = (# of nodes + 1)/2
      real alpha_save(ntr)

      if(iflag.eq.0) then ! open data file for saving
         call vzero(alpha_save,ntr)
         open(23,file='leafindex.dat',status='unknown')
         write(23,1101) ntot,nevt1,nevt2,nsig_train,nbkgd_train,
     &            ntr,node/2,iepsilon,epsilon,beta
 1101    format(8(1x,I10),2(1x,f10.5))

      elseif(iflag.eq.1) then ! fill leafindex for each tree
         write(23,1102) ((leaf_index(ileaf,k),k=1,60),ileaf=1,nleaf)
 1102    format(60e14.6)
      else ! save alpha(m) or epsilon(m)
         if(iepsilon.eq.1) then ! epsilon boosting algorithm
            do i = 1, ntr
               alpha_save(i) = epsilon
            enddo
         else ! adaboost algorithm
            do i = 1, ntr
               alpha_save(i) = alpha(i)
            enddo
         endif

         write(23,1103) (alpha_save(itr),itr=1,ntr)
 1103    format(10e14.6)
         close(23)
      endif

*                                              end save_leafindex
      end
****************************************************************************
      subroutine save_treeindex(tree_subindex)
* save all nodes information needed to test new MC 
* Roe's format
****************************************************************************
      include 'boost.inc'

      real tree_subindex(ntr,node,12)
      real purity_left(node,ntr), purity_right(node,ntr)
      integer next_left_node(node,ntr), next_right_node(node,ntr)
      real node_val(node,ntr)
      integer node_var(node,ntr)

*---- initialization

      call vzero(purity_left,node*ntr)
      call vzero(purity_right,node*ntr)
      call vzero(next_right_node,node*ntr)
      call vzero(next_left_node,node*ntr)
      call vzero(node_var,node*ntr)
      call vzero(node_val,node*ntr)

*---- open file for saving

      open(25,file='treeindex.dat',status='unknown')

      ratio = 0.1 ! arbitrary number here
      sigeff = 0.8 ! arbitrary number here

      if(iepsilon.eq.1) then
         ada_or_epsilon = -1
      else
         ada_or_epsilon = 1
      endif

      do itr = 1, ntr
         do inode = 1, node-1
            if(tree_subindex(itr,inode,1).eq.1) then
               purity_left(inode,itr) = 
     &            tree_subindex(itr,inode,6)/tree_subindex(itr,inode,7)
               purity_right(inode,itr) = 
     &            tree_subindex(itr,inode,8)/tree_subindex(itr,inode,9)

             next_left_node(inode,itr)=nint(tree_subindex(itr,inode,4))
             next_right_node(inode,itr)=nint(tree_subindex(itr,inode,5))
             node_var(inode,itr) = nint(tree_subindex(itr,inode,2))
             node_val(inode,itr) = tree_subindex(itr,inode,3)
          endif
       enddo
      enddo

      write(25,8800) ratio,sigeff,ntr,nevt1,nevt2,nsig_train,
     & nbkgd_train,node-1,ntr,epsilon,beta,psig,ada_or_epsilon
      
      write(25,8801) (alpha(itr),itr=1,ntr)

      do itr = 1, ntr
         do inode = 1, node-1
            write(25,8802) 
     &        purity_right(inode,itr),purity_left(inode,itr),
     &        next_right_node(inode,itr),next_left_node(inode,itr),
     &        node_var(inode,itr),node_val(inode,itr)
         enddo
      enddo
      
 8800 format(2(1x,f10.2),7(1x,I10),4(1x,f10.5))
 8801 format(10(1x,E12.5))
 8802 format(2(1x,E12.5),3(1x,I12),(1x,E12.5))
      close(25)

*                                                    end save_treeindex
      end
****************************************************************************
      subroutine check_variable(tree_subindex,gini_index_all)
****************************************************************************
* check irrelevant variables for boosting algorithm
* to calculate how many times input variables used for splitters
****************************************************************************
      include 'boost.inc'
      include 'hbookboost.inc'

      real tree_subindex(ntr,node,12)
      real leaf_index(node/2,2*node) ! # of leaves = (# of nodes + 1)/2
      integer nc_var(nvar-2), nc_var_tot
      real gini_index_all(ntr,node)
      real nc_gini(nvar-2), nc_gini_tot

      open(21,file='check_variable.dat',status='unknown')
      write(21,*) '==> To check irrelevant variables <=='

      call vzero(nc_var,nvar-2)
      call vzero(nc_gini,nvar-2)
      nc_var_tot = 0
      nc_gini_tot = 0.

      do itr = 1, ntr
         do inode = 1, node-1
*            write(21,*) (tree_subindex(itr,inode,k),k=1,10)
            if(nint(tree_subindex(itr,inode,1)).eq.1) then
               numvar = nint(tree_subindex(itr,inode,2))
               nc_var(numvar) = nc_var(numvar) + 1
               nc_var_tot = nc_var_tot + 1
               nc_gini(numvar)=nc_gini(numvar)+gini_index_all(itr,inode)
               nc_gini_tot=nc_gini_tot+gini_index_all(itr,inode)
            endif
         enddo
         do i = 1, nvar-2
            array2(i) = nc_var(i)
         enddo
         array2(nvar-1) = itr
         call hfn(11,array2)
      enddo
      
      write(21,*) 'Total number of splitter is', nc_var_tot
      write(21,*) 'Var# Splitter# ratio(%) gini_index  ratio(%)'

      do i = 1, nvar-2 
         write(21,411) i,nc_var(i),100.*nc_var(i)/real(nc_var_tot)
     &                 ,nc_gini(i),100.*nc_gini(i)/nc_gini_tot
      enddo
 411  format(2(1x,I6),3(1x,f10.2))
*                                                   end check_variable
      end
****************************************************************************
      subroutine update_weight(itr,leaf_index)
****************************************************************************
      include 'boost.inc'

      real leaf_index(node/2,2*node) ! # of leaves = (# of nodes + 1)/2
      real ierr(ntrain,3)
      integer itr

      call vzero(ierr,ntrain*3)
*--- 
      do i = 1, ntrain
         do ileaf = 1, nleaf
            icheck = 0
            do jcut = 1, leaf_index(ileaf,1)
               vnum = leaf_index(ileaf,3*jcut+1)
               vcut = leaf_index(ileaf,3*jcut+2)
               vlr  = leaf_index(ileaf,3*jcut+3)
               
               if(vlr.gt.998.and.dat_train(vnum,i).ge.vcut) then ! right node (great than vcut)
                  icheck = icheck + 1
               elseif(vlr.lt.-998.and.dat_train(vnum,i).lt.vcut) then ! left node (less than vcut)
                  icheck = icheck + 1
               endif
            enddo

            if(icheck.eq.leaf_index(ileaf,1)) then ! find correct leave for this event
               ierr(i,1) = ileaf ! #i data falling in #ileaf leaf
               ierr(i,2) = dat_train(nvar,i) ! signal or background	       
            endif
         enddo
      enddo

*--- compute err(m) and alpha(m) -- for adaboost algorithm
         errm = 0.
         wt_tot = 0.
         errx = 0.
         do j = 1, ntrain
            ileaf = ierr(j,1)
            purityS=leaf_index(ileaf,2)/leaf_index(ileaf,3)
            if(purityS.gt.psig) then ! all events in this leaf are signal
               isig = 1
            elseif(purityS.le.psig) then
               isig = 0
            endif
            if(nint(ierr(j,2)).eq.isig) then
               ierr(j,3) = -1.
            else
               ierr(j,3) = 1.
               errm = errm + ierr(j,3)*wt(j) ! compute error
               errx = errx + 1.
            endif	
            wt_tot = wt_tot + wt(j)
         enddo

         errm = errm/wt_tot

         print *, 'err rate = ',errx/(real(ntrain))

         if(errm.eq.0) then
            print *,'errm is zero, all distinguished'
         else                   !  beta is a tunning parameter for alpha
            alpha(itr) = beta*log((1.-errm)/errm)
         endif

*---- re-normalize weight of each trainning events
         wt_tot = 0.

         do j = 1, ntrain
            if(nint(ierr(j,3)).eq.1) then
               if(iepsilon.eq.1) then
                  wt(j) = wt(j)*exp(2*epsilon)  ! using epsilon-boosting
               else
                  wt(j) = wt(j)*exp(alpha(itr)) ! using adaBoost algorithm
               endif
            endif
            wt_tot = wt_tot + wt(j)
         enddo

         print *, 'weight total = ', wt_tot
         do j = 1, ntrain
            wt(j) = wt(j)*ntrain/wt_tot
         enddo
*---                                                      end update_weight
      end
****************************************************************************
      subroutine check_purity(itr,tree_combined)
* to estimate double ratio vs. signal efficiency and to save in ntuples
****************************************************************************
      include 'boost.inc'
      include 'hbookboost.inc'
      integer itr
      real tree_combined(nevt)
      real treecomb_tmp, prob_sig

      if(iepsilon.eq.1) then  ! epsilon boosting algorithm
         factor = 1.
         step = 2.
      else ! adaboosting algorithm
         factor = alpha_tot/real(itr)
         step = 1.
      endif

      do pt = 0, itr, step  ! for different signal efficiency
         sig_tmp = 0.
         sig_tot = 0.
         tot_tmp = 0.
           
         do i = 1, nevt_test
            if(tree_combined(i).ge.pt*factor) then
               tot_tmp = tot_tmp + 1.
               if(i.le.nsig_test) then
                  sig_tmp = sig_tmp + 1.
               endif
            endif
         enddo

*---- compute eff, purity, error rate and ratio (Ns/sqrt(N_bg))

           eff = 100.*sig_tmp/real(nsig_test) ! fraction of signal kept
           purity = 100.*sig_tmp/tot_tmp
           effbg = real(tot_tmp-sig_tmp)/real(nbkgd_test) ! fraction of bkgd kept
           effratio = 100*100.*effbg/eff

           if(eff.gt.0.) then
              print *, 'tree#/eff/purity/err_ratio (%)=',itr,eff,purity,
     &              effratio
           endif

           array(1) = itr
           array(2) = eff
           array(3) = purity
           array(4) = effratio
           call hfn(10,array)
      enddo      

*---- save boosting output for each 100 tree iterations
      if(mod(itr,100).eq.0) then
         do i = 1, nevt_test
            if(iepsilon.eq.1) then ! epsilon boost
               treecomb_tmp = tree_combined(i)*epsilon
            else ! adaboost
               treecomb_tmp = tree_combined(i)
            endif
	    
c            prob_sig = 1./(1.+exp(-2.*treecomb_tmp)) ! signal probability estimation
	    prob_sig = treecomb_tmp ! boosting output for test event

            array3(1) = itr
            array3(2) = dat_test(50,i) ! energy
            array3(3) = dat_test(51,i) ! radius to tank center
            array3(4) = dat_test(52,i) ! r to tank wall
            array3(5) = prob_sig  ! boosting output for test event
            array3(6) = dat_test(nvar,i) ! 1 for signal, <=0 for background
            call hfn(12,array3)
         enddo
      endif
      
c                                                           end check_purity
      end


****************************************************************************
****************************************************************************
      SUBROUTINE PLOT(IFLAG)
****************************************************************************
      include 'hbookboost.inc'

      INTEGER IFLAG
      INTEGER IFILE

*---> book ntpl/hist
      IF(IFLAG.EQ.1) THEN
c         CALL HLIMIT(NWPAWC)
         CALL HCDIR('//PAWC',' ')
         call hropen(2,'RWN','boost.ntp','N',1024,ISTAT)
         call hbookn(10,'check purity',NVAR1,'RWN',1024,CH)
         call hbookn(11,'spliting variables',NVAR2,'RWN',1024,CH2)
         call hbookn(12,'boosting output',NVAR3,'RWN',1024,CH3)
      ENDIF

*---> fill
      IF(IFLAG.EQ.2) THEN

      ENDIF

*---> close ntpl
      IF(IFLAG.EQ.3) THEN
        call hcdir('//RWN',' ')
        call hrout(0,icycle,' ')
        call hrend('RWN')
      ENDIF

*                                           END PLOT
      END
****************************************************************************
      subroutine tree_creator(datnode,subindex,leaf_index)
****************************************************************************
* create one binary tree(gini index is used as criteria for next split) 
****************************************************************************
      include 'boost.inc'

      real var(nvar,3)      ! 1 for minimum, 2 for maximum, 3 for step
      integer i, j, k
      integer nevtin, iflag
      real subindex(node,12)
      integer node_visited(node)
      integer knode, k2node
      integer nsbg, nsig_tmp,ntot_tmp
      integer nodeevt_index(node)
      real pgini
      real leaf_index(node/2,2*node)
c                      |   |--> 1-# of cut,2-Ns,3-Ntot,(4-var#,5-cut,6-L/R)
c                      |------> leave # 

*--- initilization
      call vzero(subindex,node*10)
      call vzero(subnode_index,node*ntrain)
      call vzero(node_visited,node)
      call vzero(leaf_index,node*node)
      call vzero(nodeevt_index,node)

      knode = 1
      k2node = 1
      nevtin = ntrain
      idebug = 0

881    continue

      if(idebug.eq.1) print *, 'node_visited = ', node_visited

*--- determine min/max of variables
      call var_min_max(datnode,nevtin,var,idebug)

*--- find subnode
      call node_finder(knode,k2node,datnode,nevtin,
     &	               var,subindex,subnode_index,pgini,idebug)


      gini_index(knode) = pgini

      node_visited(knode) = 1

      if(subindex(knode,1).gt.0.99) then
	k2node = k2node + 2 ! total nodes
        jnodeL = subindex(knode,4)
        jnodeR = subindex(knode,5)
      else
        print *, 'warning!!! no subnode found ',knode
        node_visited(knode) = 0 ! terminal node - leaf
        gini_index(knode) = -1000.
        goto 882
      endif	

      purityL = subindex(knode,6)/real(subindex(knode,7))
      purityR = subindex(knode,8)/real(subindex(knode,9))
      psigL = purityL*(1.-purityL)
      psigR = purityR*(1.-purityR)
      ntotL = subindex(knode,11)
      ntotR = subindex(knode,12)

      nodeevt_index(jnodeL) = ntotL
      nodeevt_index(jnodeR) = ntotR

      if(idebug.ge.1) then
         print *, 'purity(psig) L/R = ', purityL, purityR, psigL, psigR
         print *, 'total L/R = ', subindex(knode,7),subindex(knode,9)
      endif

*--- check gini index of left node
      call vzero(datnode,nvar*ntrain)
      do i = 1, ntotL
         do j = 1, nvar
            datnode(j,i) = dat_train(j,subnode_index(jnodeL,i))
         enddo
      enddo
      knode = jnodeL
      nevtin = ntotL
      call node_finder(knode,k2node,datnode,nevtin,
     &     var,subindex,subnode_index,pgini,idebug)

      gini_index(knode) = pgini

      if(k2node.ge.node-1) goto 888  ! end of tree

*--- check gini index of right node      
      call vzero(datnode,nvar*ntrain)
      do i = 1, ntotR
         do j = 1, nvar
            datnode(j,i) = dat_train(j,subnode_index(jnodeR,i))
         enddo
      enddo
      knode = jnodeR
      nevtin = ntotR
      call node_finder(knode,k2node,datnode,nevtin,
     &     var,subindex,subnode_index,pgini,idebug)
      gini_index(knode) = pgini

*--- select which node for next split
 882  gini_check = -999.

      do i = 1, k2node
         if(idebug.eq.1) print *, 'node_visited/gini_index(',i,')=',
     &             node_visited(i),gini_index(i)
         if(node_visited(i).ne.1.and.gini_index(i).gt.gini_check) then
            knode = i
            gini_check = gini_index(i)
         endif
      enddo

      if(idebug.eq.1) then
         print *, 'node ',knode, ' is selected for next split'
      endif

*--- refill data for next split

      call vzero(datnode,nvar*ntrain)
      ntot_tmp = nodeevt_index(knode)

      do i = 1, ntot_tmp
         do j = 1, nvar
            datnode(j,i) = dat_train(j,subnode_index(knode,i))
         enddo
      enddo
      nevtin = ntot_tmp

      goto 881                  ! return for next split

*--- tree is done
 888  continue

      if(idebug.eq.0) print *, 'node_visited = ', node_visited

*--- All subnode are found
*--- extract leaf_index information
c      real leaf_index(node,node)
c                      |   |--> 1-# of cut,2-Ns,3-Ntot,(4-var#,5-cut,6-L/R)
c                      |------> leave #
cccc

      nleaf = 0
      print *, ' =======>    Tree is done     <====== '
      print *, ' =======>   Node index info   <====== '
      do i = 1, node-1
c        if(subindex(i,1).lt.0.99.and.subindex(i,10).gt.0) then
         if(node_visited(i).eq.0) then ! it's terminal node - leave
          subindex(i,1) = 0.
          nleaf = nleaf + 1
          nvcut = 0

          lnode = i
311       knode_father = subindex(lnode,10)
          if(knode_father.ge.1) then
            nvcut = nvcut + 1
            lnode = knode_father
            goto 311
          endif

          if(idebug.eq.0) then
             print *,'#',i,(subindex(i,j),j=1,10),' leaf # ',nleaf
             print *, '==> # of cuts: ',nvcut
          endif

          leaf_index(nleaf,1) = nvcut
          lnode = i
312       knode_father = subindex(lnode,10)
          if(knode_father.ge.1) then
            if(subindex(knode_father,4).eq.i) then ! left node
              nsig_tmp = subindex(knode_father,6)
              ntot_tmp = subindex(knode_father,7)
              leaf_index(nleaf,2) = nsig_tmp
              leaf_index(nleaf,3) = ntot_tmp
            elseif(subindex(knode_father,5).eq.i) then ! right node
              nsig_tmp = subindex(knode_father,8)
              ntot_tmp = subindex(knode_father,9)
              leaf_index(nleaf,2) = nsig_tmp
              leaf_index(nleaf,3) = ntot_tmp
            endif

            if(subindex(knode_father,4).eq.lnode) then
               leaf_index(nleaf,nvcut*3+3) = -999.
            elseif(subindex(knode_father,5).eq.lnode) then
               leaf_index(nleaf,nvcut*3+3) = 999.
            endif

            leaf_index(nleaf,nvcut*3+1)=subindex(knode_father,2)
            leaf_index(nleaf,nvcut*3+2)=subindex(knode_father,3)
            nvcut = nvcut - 1
            lnode = knode_father
            goto 312
          endif

        endif
      enddo

*--- # of leaves and nodes found
      print *, '==============================================='
      print *, ' Totally ', nleaf, ' leaves found !!! '
      print *, ' Totally nodes are ', 2*nleaf-1
      print *, '==============================================='

*---                                               end tree_creator
      end
***************************************************************************
      subroutine tree_creator2(datnode,subindex,leaf_index)
****************************************************************************
* create one binary tree(no gini index is used, leaves # varies)
* left node is always selected as next splitter
****************************************************************************
      include 'boost.inc'

      real var(nvar,3)      ! 1 for minimum, 2 for maximum, 3 for step
      integer i, j, k
      integer nevtin, iflag
      real subindex(node,12)
      integer node_visited(node)
      integer knode, k2node
      integer nsbg, nsig_tmp
      real pgini
      real leaf_index(node/2,2*node)
c                      |   |--> 1-# of cut,2-Ns,3-Ntot,(4-var#,5-cut,6-L/R)
c                      |------> leave #

*--- initilization
      call vzero(subindex,node*10)
      call vzero(subnode_index,node*ntrain)
      call vzero(node_visited,node)
      call vzero(leaf_index,node*node)

      knode = 1
      k2node = 1
      nevtin = ntrain
      idebug = 0

11    continue
      if(k2node.ge.node-1) goto 993
      print *, ' '
      if(idebug.eq.1) print *, 'node_visited = ', node_visited

*--- determine min/max of variables
      call var_min_max(datnode,nevtin,var,idebug)

*--- find subnode
      call node_finder(knode,k2node,datnode,nevtin,
     &                 var,subindex,subnode_index,pgini,idebug)

      node_visited(knode) = 1
      if(subindex(knode,1).gt.0.99) then
        k2node = k2node + 2 ! total nodes
        jnodeL = subindex(knode,4)
        jnodeR = subindex(knode,5)
      else
        jnodeL = knode
        jnodeR = knode
        print *, 'one leaf found ',knode
        goto 991
      endif


      purityL = subindex(knode,6)/real(subindex(knode,7))
      purityR = subindex(knode,8)/real(subindex(knode,9))
      psigL = purityL*(1.-purityL)
      psigR = purityR*(1.-purityR)
      ntotL = subindex(knode,11)
      ntotR = subindex(knode,12)
      nsigL = min(subindex(knode,6),ntotL-subindex(knode,6))
      nsigR = min(subindex(knode,8),ntotR-subindex(knode,8))

      if(ntotL.eq.0) then
        psigL = 0.
        purityL = 0.
      endif

      if(ntotR.eq.0) then
        psigR = 0.
        purityR = 0.
      endif

      if(idebug.ge.0) then
         print *, 'purity(psig) L/R = ', purityL, purityR, psigL, psigR
      endif

** to find left leaf

      if(psigL.lt.psig0.or.ntotL.lt.nmax) then ! one leaf found
        print *, 'one left leaf found ',jnodeL
        node_visited(jnodeL) = 1
      else
        call vzero(datnode,nvar*ntrain)
        do i = 1, ntotL
          do j = 1, nvar
            datnode(j,i) = dat_train(j,subnode_index(jnodeL,i))
          enddo
        enddo
        knode = jnodeL
        nevtin = ntotL
        print *, '--> try left node ',knode
        goto 11
      endif

** to find right leaf

      if(psigR.lt.psig0.or.ntotR.lt.nmax) then ! one leaf found
        print *, 'one right leaf found ',jnodeR
        node_visited(jnodeR) = 1
      else
        call vzero(datnode,nvar*ntrain)
        do i = 1, ntotR
          do j = 1, nvar
            datnode(j,i) = dat_train(j,subnode_index(jnodeR,i))
          enddo
        enddo
        knode = jnodeR
        nevtin = ntotR
        print *, '--> try right node ',knode
        goto 11
      endif

** if some nodes are still not visited, try it now
991   continue
      knode = 0
      print *, ' '
      print *, '**** k2node = ',k2node
      do i = k2node, 1, -1
        if(node_visited(i).ne.1) then
          print *, '==> Non-visited node is', i
          knode = i
          goto 992
        endif
      enddo
992   continue

      if(knode.gt.1.and.knode.le.k2node) then
        print *, '==> try non-visited node ',knode
        call vzero(datnode,nvar*ntrain)
        knode_father = subindex(knode,10)
        if(knode.eq.subindex(knode_father,4)) then
          print *, 'node',knode,' is left of node',knode_father
          ntotLR = subindex(knode_father,11)
          nsig_tmp   = subindex(knode_father,6)
          nsbg   = min(nsig_tmp,(ntotLR-nsig_tmp))
          nevtin = ntotLR
          if(ntotLR.le.nmax) then
            print *, 'One left leaf found ',knode
            node_visited(knode) = 1
            goto 991
          endif

        elseif(knode.eq.subindex(knode_father,5)) then
          print *, 'node',knode,' is right of node',knode_father
          ntotLR = subindex(knode_father,12)
          nsig_tmp   = subindex(knode_father,8)
          nsbg   = min(nsig_tmp,(ntotLR-nsig_tmp))
          nevtin = ntotLR
          if(ntotLR.le.nmax) then
            print *, 'One right leaf found ',knode
            node_visited(knode) = 1
            goto 991
          endif
        else
          print *, ' XXXX Wrong, Can not find correct node XXXX'
          print *, ' XXXX ',knode,knode_father
        endif

        if(k2node.ge.node-1) then
           node_visited(knode) = 1
           goto 991
        endif

*--- refill knode for next split

        do i = 1, ntotLR
          do j = 1, nvar
            datnode(j,i) = dat_train(j,subnode_index(knode,i))
          enddo
        enddo

        goto 11
      endif

*--- All subnode are found
*--- extract leaf_index information
c      real leaf_index(node,node)
c                      |   |--> 1-# of cut,2-Ns,3-Ntot,(4-var#,5-cut,6-L/R)
c                      |------> leave #
cccc
 993  continue
      nleaf = 0
      print *, ' =======>    Tree is done     <====== '
      print *, ' =======>   Node index info   <====== '
      do i = 1, node
        if(subindex(i,1).lt.0.99.and.subindex(i,10).gt.0) then
          nleaf = nleaf + 1
          nvcut = 0

          lnode = i
311       knode_father = subindex(lnode,10)
          if(knode_father.ge.1) then
            nvcut = nvcut + 1
            lnode = knode_father
            goto 311
          endif

          if(idebug.eq.1) then
             print *,'#',i,(subindex(i,j),j=1,10),' leaf # ',nleaf
             print *, '==> # of cuts: ',nvcut
          endif

          leaf_index(nleaf,1) = nvcut
          lnode = i
312       knode_father = subindex(lnode,10)
          if(knode_father.ge.1) then
            if(subindex(knode_father,4).eq.i) then
              nsig_tmp = subindex(knode_father,6)
              ntot_tmp = subindex(knode_father,7)
              leaf_index(nleaf,2) = nsig_tmp
              leaf_index(nleaf,3) = ntot_tmp
            elseif(subindex(knode_father,5).eq.i) then
              nsig_tmp = subindex(knode_father,8)
              ntot_tmp = subindex(knode_father,9)
              leaf_index(nleaf,2) = nsig_tmp
              leaf_index(nleaf,3) = ntot_tmp
            endif

            if(subindex(knode_father,4).eq.lnode) then
               leaf_index(nleaf,nvcut*3+3) = -999.
            elseif(subindex(knode_father,5).eq.lnode) then
               leaf_index(nleaf,nvcut*3+3) = 999.
            endif

            leaf_index(nleaf,nvcut*3+1)=subindex(knode_father,2)
            leaf_index(nleaf,nvcut*3+2)=subindex(knode_father,3)
            nvcut = nvcut - 1
            lnode = knode_father
            goto 312
          endif

        endif
      enddo

*--- # of leaves and nodes found
      print *, '==============================================='
      print *, ' Totally ', nleaf, ' leaves found !!! '
      print *, ' Totally nodes are ', 2*nleaf-1
      print *, '==============================================='

*                                                      end tree_creator2
      end
****************************************************************************
      subroutine readin_dat(datnode,idebug)
****************************************************************************
* input data file
* idebug - idebug = 1 for debug
* output - datnode
****************************************************************************
      include 'boost.inc'

      integer i, j, k, nevt1, nevt2
      integer idebug
      real dat0(nvar,nevt)
      integer kevt_selected(ntrain)
      integer kevt_overlapped(ntrain)
c      double precision rand(ntrain)
      real rand(ntrain)

      call vzero(dat0,nvar*nevt)
      call vzero(dat_train,nvar*ntrain)
      call vzero(dat_test,nvar*nevt)
      call vzero(datnode,nvar*ntrain)
      call vzero(kevt_selected,ntrain)
      call vzero(kevt_overlapped,ntrain)
      call vzero(rand,ntrain)

*--- readin data file
 
      open(21,file=dfile,status='old')
      do i = 1, nevt
         if(i.eq.1) read(21,*) nevt1, nevt2
         read(21,*,end=100) (dat0(ivar,i),ivar=1,nvar-2)                  
 
         dat0(nvar-1,i) = 0
         if(i.le.nevt1) then ! signal
            dat0(nvar,i) = 1 
         elseif(i.gt.nevt1) then ! background
            dat0(nvar,i) = 0
         endif            
      enddo
 100  close(21)
      ntot = i - 1  

      if(irand.eq.1) then ! select training sample randomly
cc
*--- select ntrain events from signal and background for trainning

      CALL RANMAR(RAND,NTRAIN) ! real
c      CALL RM48(RAND,NTRAIN)   ! double precision

*--- ! selected signal events for trainning
      koverlap = 0  ! overlapped events

      do i = 1, nsig_train  ! selected signal events for trainning
         kevt = 1 + nint(rand(i)*nevt1)
         if(kevt.gt.nevt1) kevt = nevt1
         if(kevt.le.1) kevt = 1

	 if(dat0(nvar-1,kevt).eq.1) then
	   koverlap = koverlap + 1
	   kevt_overlapped(koverlap) = i  ! event i will be replaced later
	 endif

         do j = 1, nvar
            dat_train(j,i) = dat0(j,kevt)
         enddo
         dat_train(nvar-1,i) = i
         dat0(nvar-1,kevt) = 1 ! means selected for trainning
         kevt_selected(i) = kevt ! for event index
      enddo

*--- if overlapped events found, replace them
      print  *, 'overlapped events for signal =',koverlap
      do k = 1, koverlap
        kevt_needreplace = kevt_overlapped(k)
        do i = 1, nevt1 
	   if(dat0(nvar-1,i).eq.0) then
	      do j = 1, nvar
		 dat_train(j,kevt_needreplace) = dat0(j,i)
              enddo
	      dat0(nvar-1,i) = 1 ! event selected for trainning
              dat_train(nvar-1,kevt_needreplace)=kevt_needreplace
              goto 731
	   endif
        enddo
731     continue
      enddo

*--- ! select background events for trainning
      koverlap = 0  ! overlapped events
      call vzero(kevt_overlapped,ntrain)

       do i = nsig_train+1, ntrain  ! select background events for trainning
         kevt = nevt1 + 1 + nint(rand(i)*nevt2)
         if(kevt.gt.ntot) kevt = ntot
         if(kevt.le.nevt1) kevt = nevt1 + 1

	 if(dat0(nvar-1,kevt).eq.1) then
	   koverlap = koverlap + 1
           kevt_overlapped(koverlap) = i  ! will be replaced later
	 endif

         do j = 1, nvar
            dat_train(j,i) = dat0(j,kevt)
         enddo
         dat_train(nvar-1,i) = i
	 dat0(nvar-1,kevt) = 1 ! means selected for trainning
         kevt_selected(i) = kevt ! for event index
       enddo

c      print *, 'kevt_selected'
c      print *, kevt_selected

*--- if overlapped events found, replace them
       print  *, 'overlapped events for background =',koverlap
      
       do k = 1, koverlap
         kevt_needreplace = kevt_overlapped(k)
         do i = nevt1+1, ntot
           if(dat0(nvar-1,i).eq.0) then
              do j = 1, nvar
                 dat_train(j,kevt_needreplace) = dat0(j,i)
              enddo
              dat0(nvar-1,i) = 1 ! event selected for trainning
              dat_train(nvar-1,kevt_needreplace)=kevt_needreplace
              goto 732
           endif
         enddo
732      continue
       enddo

*--- fill events for test (exclude trainning events)

      nevt_test  = 0 ! total events for test
      nsig_test  = 0 ! total signal for test
      nbkgd_test = 0 ! total background for test

      do i = 1, ntot
         if(i.eq.nevt1+1) nsig_test = nevt_test  ! total signal for test
         if(dat0(nvar-1,i).eq.0) then ! not selected
            nevt_test = nevt_test + 1
            do j = 1, nvar
               dat_test(j,nevt_test) = dat0(j,i)
            enddo
            dat_test(nvar-1,nevt_test) = nevt_test ! for event index
         endif
      enddo

      nbkgd_test = nevt_test - nsig_test

*---->>>
      else ! select training sample sequentially
        nevt_train = 0 ! total events for training
        nevt_test  = 0 ! total events for test
        nsig_test  = 0 ! total signal for test
        nbkgd_test = 0 ! total background for test

        do i = 1, ntot
          if(i.le.nsig_train.or.
     &          i.gt.nevt1.and.i.le.nevt1+nbkgd_train) then ! for training
            nevt_train = nevt_train + 1
            do j = 1, nvar
	      dat_train(j,nevt_train) = dat0(j,i)
	    enddo
            dat_train(nvar-1,nevt_train) = nevt_train
          else ! for test sample
            nevt_test = nevt_test + 1
            do j = 1, nvar
              dat_test(j,nevt_test) = dat0(j,i)
            enddo
            dat_test(nvar-1,nevt_test) = nevt_test
          endif
        enddo

	nsig_test = nevt1 - nsig_train
        nbkgd_test = nevt2 - nbkgd_train

      endif
*---->>>
*---- fill trainning events in root node of binary tree
      do i = 1, ntrain
         do j = 1, nvar            
            datnode(j,i) = dat_train(j,i)
         enddo
      enddo

      print *, '==================================================='
      print *, 'total/sgnl/bkgd/ events = ', ntot, nevt1, nevt2
      print *, 's/bg events for trainning = ', nsig_train, nbkgd_train
      print *, 's/bg events for test = ',
     &     nevt_test,nsig_test,nbkgd_test
      print *, '==================================================='

*----> write test sample
      open(22,file='test.dat',status='unknown')
      write(22,*) nsig_test,nbkgd_test
      write(22,1000) 
     &     ((dat_test(ivar,ievt),ivar=1,nvar-2),ievt=1,nevt_test)

 1000 format(<nvar-2>(e14.6))
      
*                                                           end readin_dat
      end
****************************************************************************
      subroutine var_min_max(datnode,nevtin,var,idebug)
****************************************************************************
* datnode <-- input data
* nevtin  <-- number of events
* idebug  <-- idebug = 1 for debug
* var     --> output of min/max of variables
****************************************************************************
      include 'boost.inc'

      real var(nvar,3)          ! 1 for minimum, 2 for maximum, 3 for step
      integer i,j,nevt1,nevt2,nevtin
      real nbin(nvar,nc,3) ! 1 for Ntot, 2 for Ns, 3 for threshold cut
      real nbin2(nvar,nc,3)
      integer nsig_tot, nsig_tmp, ntot_tmp
      integer idebug

      call vzero(nbin,nvar*nc*3)
      call vzero(var,nvar*3)

*--- for debug

      if(idebug.eq.1) then
         print *, ' '
         print *, ' ===> call var_min_max'
      endif

*--- init

      do j = 1, nvar
         var(j,1) = 99999.   ! for minimum
         var(j,2) = -99999.  ! for maximum
      enddo

*--- determine min/max of variables
      do i = 1, nevtin
         do j = 1, nvar            
            if(var(j,1).gt.datnode(j,i)) var(j,1) = datnode(j,i)
            if(var(j,2).lt.datnode(j,i)) var(j,2) = datnode(j,i)
         enddo
      enddo

      do j = 1, nvar
         var(j,3) = (var(j,2) - var(j,1))/real(nc)
      enddo

*--- re-define min/max if events in left/right side is less than nmin
ccc locate events in variables' bin / used for relocation of var_min_max

      do iteration = 1, 1

         call event_bin(datnode,nevtin,var,nbin,nbin2,idebug)

         nsig_tot = 0
         do ibin = 1, nc
            nsig_tot = nsig_tot + nbin2(1,ibin,2)
         enddo
         if(idebug.eq.1) print *, 'tot_signal = ', nsig_tot

*  re-determine var_min_max values from left and right side
         do j = 1, nvar-2 
            ntot_tmp = 0
            do ibin = 1, nc-1
               ntot_tmp = ntot_tmp + nbin2(j,ibin,1)
               if(ntot_tmp.le.nmin) then 
                  var(j,1) = var(j,1) + var(j,3) 
               endif
            enddo
               
            ntot_tmp = 0
            do ibin = nc, 1, -1
               ntot_tmp = ntot_tmp + nbin2(j,ibin,1)
               if(ntot_tmp.le.nmin) then
                  var(j,2) = var(j,2) - var(j,3)
               endif
            enddo
         enddo

         if(idebug.eq.1) print *, '===> after bin re-allocation <==='
         
         do j = 1, nvar
            var(j,3) = (var(j,2) - var(j,1))/real(nc)
         enddo
      enddo

c                                                 end var_min_max
      end
****************************************************************************
      subroutine event_bin(datnode,nevtin,var,nbin,nbin2,idebug)
****************************************************************************
* datanode  <-- input dat information
* nevtin    <-- # of events
* var       <-- min/max values of variables
* nbin      --> output # of events in each bin
* idebug    <-- idebug = 1 for debug
****************************************************************************
      include 'boost.inc'

      real var(nvar,3)          ! 1 for minimum, 2 for maximum, 3 for step
      integer i,j,nevt1,nevt2,nevtin
      real nbin(nvar,nc,3) ! 1 for Ntot, 2 for Ns, 3 for threshold cut (weighted)
      real nbin2(nvar,nc,3) ! for true event information
      integer idebug
      real wt_tmp

      call vzero(nbin,nvar*nc*3)
      call vzero(nbin2,nvar*nc*3)

ccc locate events in variables' bin

      do j = 1, nvar-2
         do i = 1, nc
            nbin(j,i,3) = var(j,1) + real(i)*var(j,3)
         enddo

         do i = 1, nevtin
            wt_tmp = 1.
	    ievt = nint(datnode(nvar-1,i))
            wt_tmp = wt(ievt)

            ibin = int((datnode(j,i)-var(j,1))/var(j,3)) + 1
            if(ibin.lt.1) then
               ibin = 1
            elseif(ibin.gt.nc) then
               ibin = nc
            endif
            nbin2(j,ibin,1) = nbin2(j,ibin,1) + 1   ! for true event info
            nbin(j,ibin,1) = nbin(j,ibin,1) + wt_tmp ! for weighted event info
            if(datnode(nvar,i).gt.0.99) then ! if(signal, fill in nbin(nvar,nc,2)
               nbin2(j,ibin,2) = nbin2(j,ibin,2) + 1
               nbin(j,ibin,2) = nbin(j,ibin,2) + wt_tmp
            endif
         enddo

         if(idebug.eq.1) then
            print *, 'nbin(',j,',*,1)=',(nbin(j,jbin,1),jbin=1,nc)
            print *, 'nbin(',j,',*,2)=',(nbin(j,jbin,2),jbin=1,nc)         
         endif
      enddo

c                                                        end event_bin
      end
****************************************************************************
      subroutine node_finder(knode,k2node,
     &datnode,nevtin,var,subindex,subnode_index,pgini,idebug)
****************************************************************************
* knode  - # of current node, k2node - max node number used
* datnode  - dat in current node
* nevtin - number of events in father node
* var    - min/max values of variables
* output - 
*     subnode_index(#node,evt) - subnode event # information
*     subindex(node,12) --> found(1)/0, 2-x_j, 3-threshold, 4-son left, 5 son-right,
*                          6-Ns-l, 7-Ntot-l, 8-Ns-r, 9-Ntot-r, 10-father node
*                          11-total events in left node, 12-total events in right node
****************************************************************************
      include 'boost.inc'

      integer nevtin,iflag ! iflag=1 find two subnode
      real var(nvar,3)
      real subindex(node,12) ! 1/0, x_j, threshold, left son(node), right son(node),
                      ! Ns_l, Ns+Nb_l, Ns_r, Ns+Nb_r, father node
      real nbin(nvar,nc,3) ! 1 for Ntot, 2 for Ns, 3 for threshold cut
      real nbin2(nvar,nc,3)
      real psig1, psig2, psig_tot, pgini
      integer nsig_tot, nsig_tmp, ntot_tmp, ntotin
      integer idebug, knode ! current node number
      integer k1, k2, numvar
      integer i,j

      call vzero(nbin,nvar*nc*3)
      call vzero(nbin2,nvar*nc*3)

*--- for debug

      if(idebug.eq.1) then
         print *, ' '
         print *, ' ===> call node_finder'
      endif

      iflag = 0

ccc locate events in variables' bin
      call event_bin(datnode,nevtin,var,nbin,nbin2,idebug)

      numvar = 0
      threshold = 0. 
      pcrit = 1.e20 ! minimize criteria
      pgini = 0.

      sig_tot = 0.
      totin = 0.
      nsig_tot = 0.
      ntotin = 0

      do ibin = 1, nc
         sig_tot  = sig_tot + nbin(1,ibin,2)
         totin    = totin + nbin(1,ibin,1)  ! weighted event info
         nsig_tot = nsig_tot + nbin2(1,ibin,2)
         ntotin   = ntotin + nbin2(1,ibin,1) ! true event info
      enddo

c      print *, 'total events/signal = ', ntotin, nsig_tot
      icheck = 0 ! for clean bkgd/signal seperation
      do j = 1, nvar-2
         tot_tmp = 0.
         sig_tmp = 0.
         ntot_tmp = 0
         nsig_tmp = 0
         do ibin = 1, nc-1
            tot_tmp = tot_tmp + nbin(j,ibin,1)
            sig_tmp = sig_tmp + nbin(j,ibin,2)
            ntot_tmp = ntot_tmp + nbin2(j,ibin,1)
            nsig_tmp = nsig_tmp + nbin2(j,ibin,2)
            if(ntot_tmp.ge.nmin.and.ntotin-ntot_tmp.ge.nmin) then
               psig1 = sig_tmp/tot_tmp
               psig2 = (sig_tot - sig_tmp)/(totin-tot_tmp)

               psum = real(tot_tmp)*psig1*(1.-psig1) + 
     &                real(totin-tot_tmp)*psig2*(1.-psig2)


               if(psum.lt.pcrit) then ! minimize criteria
                  pcrit = psum
                  numvar = j
                  vbin = ibin
                  threshold = nbin(j,ibin,3)
*--- if subnode found, fill subindex
                  subindex(knode,1) = 1
                  subindex(knode,2) = numvar
                  subindex(knode,3) = threshold
                  subindex(knode,4) = k2node+1
                  subindex(knode,5) = k2node+2
                  subindex(knode,6) = sig_tmp
                  subindex(knode,7) = tot_tmp
                  subindex(knode,8) = sig_tot - sig_tmp
                  subindex(knode,9) = totin - tot_tmp
                  subindex(k2node+1,10) = knode
                  subindex(k2node+2,10) = knode
	 	  subindex(knode,11) = ntot_tmp
		  subindex(knode,12) = ntotin - ntot_tmp
               endif
            endif
         enddo
      enddo

*--- re-allocate events in subnode
      if(subindex(knode,1).eq.1) then
         k1 = 0
         k2 = 0

* initialization subnode_index of current found nodes

         do i = 1, ntrain
            subnode_index(k2node+1,i) = 0
            subnode_index(k2node+2,i) = 0
         enddo

         do i = 1, ntotin
            if(datnode(numvar,i).le.threshold) then
               k1 = k1 + 1
               subnode_index(k2node+1,k1)=nint(datnode(nvar-1,i))
            else
               k2 = k2 + 1
               subnode_index(k2node+2,k2)=nint(datnode(nvar-1,i)) 
            endif
         enddo

*--- calculate Gini index
         psig_tot = totin*(sig_tot/totin)*(1.-sig_tot/totin)
         pgini = psig_tot - pcrit

         if(idebug.eq.1) then
            print *, 'knode,psig_tot/pcrit=',knode,psig_tot,pcrit,pgini
         endif
      endif
*--- print out for debug
      if(idebug.eq.1) then
         print *, 'Ntot_L/Ntot_R = ',k1,k2
         print *, 'pcrit/var/bin/cut = ',pcrit,numvar,vbin,threshold
         print *, '===> subindex found <==='
         print *, 'iflag var cut sonL sonR Ns_L Ntot_L Ns_R Ntot_R'
         print *, (subindex(knode,i),i=1,10)
      endif

*                                                        end node_finder
      end
****************************************************************************
